#ifndef lint
static const char rcsid[] = "$Id: dnsget.c,v 1.10 2008/07/06 20:35:20 vixie Exp $";
#endif

/* Import. */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>

#include <assert.h>
#include <errno.h>
#include <netdb.h>
#include <resolv.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* Macros. */

#define CHK(x,e) if ((x) < 0) { \
	fprintf(stderr, "%s: %s\n", e, strerror(errno)); \
	exit(1); \
}

#define ISSET(x, in) (((in) & (1 << (x))) != 0)

/* Types. */

struct rr {
	const char *name;
	ns_type type;
	const char **fields;
	int (*get)(ns_msg *, ns_rr *, int);
};

/* Forward. */

static void usage(const char *msg, ...);
static const char *setservers(res_state res, const char *servers);
static int dnsget(const char *, const struct rr *, res_state, int, int);
static int rrget(ns_msg *, ns_rr *);

static int soa_get(ns_msg *, ns_rr *, int);
static const char *soa_fields[] = {
	"mname", "rname", "serial", "refresh", "retry",
	"expire", "minimum", NULL
};
enum {
	soa_mname, soa_rname, soa_serial, soa_refresh, soa_retry,
	soa_expire, soa_minimum
};

static int ns_get(ns_msg *, ns_rr *, int);
static const char *ns_fields[] = { "nsdname", NULL };
enum { ns_nsdname };

static int a_get(ns_msg *, ns_rr *, int);
static const char *a_fields[] = { "address", NULL };
enum { a_address };

static int aaaa_get(ns_msg *, ns_rr *, int);
static const char *aaaa_fields[] = { "address", NULL };
enum { aaaa_address };

static int mx_get(ns_msg *, ns_rr *, int);
static const char *mx_fields[] = { "preference", "exchange", NULL };
enum { mx_preference, mx_exchange };

#define cname_get ns_get
static const char *cname_fields[] = { "cname", NULL };
enum { cname_cname = ns_nsdname };

#define ptr_get ns_get
static const char *ptr_fields[] = { "ptrdname", NULL };
enum { ptr_ptrdname = ns_nsdname };

/* Data. */

static const char *progname = "amnesia";
static struct rr rrs[] = {
	{ "soa",	ns_t_soa,	soa_fields,	soa_get },
	{ "ns",		ns_t_ns,	ns_fields,	ns_get },
	{ "a",		ns_t_a,		a_fields,	a_get },
	{ "aaaa",	ns_t_aaaa,	aaaa_fields,	aaaa_get },
	{ "mx",		ns_t_mx,	mx_fields,	mx_get },
	{ "cname",	ns_t_cname,	cname_fields,	cname_get },
	{ "ptr",	ns_t_ptr,	ptr_fields,	ptr_get },
	{ NULL }
};

/* Public. */

int
main(int argc, char **argv) {
	const char *domain, *cp, *typename;
	const struct rr *rr;
	struct __res_state res;
	int ch, all = 0, fields;
	char *dot;

	if ((progname = strrchr(*argv, '/')) == NULL)
		progname = *argv;
	else
		progname++;
	memset(&res, 0, sizeof res);
	CHK(res_ninit(&res), "res_ninit");
	while ((ch = getopt(argc, argv, "adrls:")) != -1) {
		switch (ch) {
		case 'a':
			all++;
			break;
		case 'd':
			res.options |= RES_DEBUG;
			break;
		case 'r':
			res.options &= ~RES_RECURSE;
			break;
		case 'l':
#ifdef RES_DONTROUTE
			res._flags |= RES_DONTROUTE;
#endif
			break;
		case 's':
			if ((cp = setservers(&res, optarg)) != NULL)
				usage(cp);
			break;
		default:
			usage("unrecognized argument -%c", ch);
		}
	}
	argv += optind, argc -= optind;
	if (argc < 1)
		usage("missing domain argument");
	domain = *argv++; argc--;
	dot = strrchr(domain, '.');
	if (dot != NULL && dot[1] == '\0')
		*dot = '\0';
	if (argc < 1)
		usage("missing rrtype argument");
	typename = *argv++; argc--;
	for (rr = rrs; rr->name != NULL; rr++)
		if (strcasecmp(rr->name, typename) == 0)
			break;
	if (rr->name == NULL)
		usage("unrecognized rr type %s", typename);
	fields = -1;	/* all ones */
	if (argc > 0) {
		cp = *argv++; argc--;
		fields = 0;
		while (*cp != '\0') {
			const char *x = strchr(cp, ',');
			int l = (x != NULL) ? x - cp : strlen(cp);
			const char **s;

			for (s = rr->fields; *s != NULL; s++)
				if (strlen(*s) == l &&
				    strncasecmp(*s, cp, l) == 0)
					break;
			if (*s == NULL)
				usage("unrecognized field name");
			fields |= (1 << (s - rr->fields));
			cp += l;
			if (*cp == ',')
				cp++;
		}
		assert(fields != 0);
	}
	if (argc > 1)
		usage("too many arguments");
	exit(dnsget(domain, rr, &res, fields, all));
}

/* Private. */

static void
usage(const char *msg, ...) {
	va_list ap;
	struct rr *rr;
	const char **s;

	va_start(ap, msg);
	fputs("usage error: ", stderr);
	vfprintf(stderr, msg, ap);
	fputc('\n', stderr);
	va_end(ap);
	fprintf(stderr, "usage: %s [-a] [-d] [-l] [-s server[,...] "
			"name type [field[,field...]]\n",
		progname);
	for (rr = rrs; rr->name != NULL; rr++) {
		fprintf(stderr, "\t%s\t", rr->name);
		for (s = rr->fields; *s != NULL; s++)
			fprintf(stderr, "%s ", *s);
		fputc('\n', stderr);
	}

	exit(1);
}

static const char *
setservers(res_state res, const char *servers) {
	union res_sockaddr_union serverset[MAXNS];
	int nscount = 0;

	while (*servers != '\0') {
		const char *x = strchr(servers, ',');
		int l = (x != NULL) ? x - servers : strlen(servers);
		char t[sizeof "ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255"];
		union res_sockaddr_union *s = &serverset[nscount];

		if (nscount > MAXNS)
			return ("too many servers specified");
		if (l + 1 > sizeof t)
			return ("server literal is too long");
		memset(t, 0, sizeof t);
		strncpy(t, servers, l);
		memset(s, 0, sizeof *s);
		if (inet_pton(AF_INET6, t, &s->sin6.sin6_addr)) {
			s->sin6.sin6_family = AF_INET6;
			s->sin6.sin6_port = htons(NS_DEFAULTPORT);
		} else if (inet_pton(AF_INET, t, &s->sin.sin_addr)) {
			s->sin.sin_family = AF_INET;
			s->sin.sin_port = htons(NS_DEFAULTPORT);
		} else {
			return ("server must be valid INET6 or INET literal");
		}
		nscount++;
		servers += l;
		while (*servers == ',')
			servers++;
	}
	res_setservers(res, serverset, nscount);
	return (NULL);
}

static int
dnsget(const char *domain, const struct rr *rr, res_state res,
       int fields, int all)
{
	char qname[NS_MAXDNAME];
	u_char buf[NS_MAXMSG];
	ns_rcode rcode;
	ns_msg msg;
	int n, sn;

	n = res_nquery(res, domain, ns_c_in, rr->type, buf, sizeof buf);
	if (n <= 0) {
		switch (res->res_h_errno) {
		case HOST_NOT_FOUND:
			fputs("NXDOMAIN", stdout);
			break;
		case TRY_AGAIN:
			fputs("TEMPFAIL", stdout);
			break;
		case NO_RECOVERY:
			fputs("PERMFAIL", stdout);
			break;
		case NO_DATA:
			fputs("NODATA", stdout);
			break;
		default:
			printf("?%d?", res->res_h_errno);
			break;
		}
		fputc('\n', stdout);
		return (1);
	}

	CHK(ns_initparse(buf, n, &msg), "ns_initparse");

	rcode = ns_msg_getflag(msg, ns_f_rcode);
	if (rcode == ns_r_noerror &&
	    (res->options & RES_RECURSE) == 0 &&
	    ns_msg_getflag(msg, ns_f_aa) == 0)
		rcode = ns_r_notauth;
	printf("%s %d %d %d\n", p_rcode(rcode),
	       ns_msg_count(msg, ns_s_an),
	       ns_msg_count(msg, ns_s_ns),
	       ns_msg_count(msg, ns_s_ar));
	if (rcode != ns_r_noerror)
		return (1);

	strncpy(qname, domain, sizeof qname);
	qname[sizeof qname - 1] = '\0';
	for (sn = ns_s_an; sn <= (all ? ns_s_ar : ns_s_an); sn++) {
		int rrn;

		for (rrn = 0; rrn < ns_msg_count(msg, sn); rrn++) {
			int need_newline = 1;
			ns_rr nsrr;

			CHK(ns_parserr(&msg, sn, rrn, &nsrr), "ns_parserr");
			if (all)
				printf("%s %s %s ",
				       p_section(sn, ns_o_query),
				       ns_rr_name(nsrr),
				       p_type(ns_rr_type(nsrr)));
			if (sn == ns_s_an) {
				if (strcasecmp(ns_rr_name(nsrr), qname) != 0) {
					printf("? %s != %s",
					       ns_rr_name(nsrr), qname);
					n = 1;
				} else if (ns_rr_type(nsrr) == ns_t_cname) {
					n = dn_expand(ns_msg_base(msg),
						      ns_msg_end(msg),
						      ns_rr_rdata(nsrr),
						      qname, sizeof qname);
					CHK(n, "dn_expand");
					if (all)
						n = cname_get(&msg, &nsrr, -1);
					else
						need_newline = n = 0;
				} else if (ns_rr_type(nsrr) != rr->type) {
					printf("? %d", ns_rr_type(nsrr));
					n = 1;
				} else {
					n = rr->get(&msg, &nsrr, fields);
				}
			} else {
				n = rrget(&msg, &nsrr);
			}
			if (need_newline)
				putchar('\n');
			if (n != 0)
				return (n);
		}
	}
	return (0);
}

static int
rrget(ns_msg *msg, ns_rr *nsrr) {
	struct rr *rr;

	for (rr = rrs; rr->name != NULL; rr++)
		if (rr->type == ns_rr_type(*nsrr))
			return (rr->get(msg, nsrr, -1));	/* all ones */
	putchar('?');
	return (0);
}

/* Private - SOA. */

static int
soa_get(ns_msg *msg, ns_rr *rr, int fields) {
	const u_char *base = ns_msg_base(*msg);
	const u_char *end = ns_msg_end(*msg);
	const u_char *rdata = ns_rr_rdata(*rr);
	char domain[NS_MAXDNAME];
	int n, printed = 0;
	u_long l;

	/* Do SOA MNAME. */
	n = dn_expand(base, end, rdata, domain, sizeof domain);
	CHK(n, "dn_expand");
	rdata += n;
	if (ISSET(soa_mname, fields)) {
		fputs(domain, stdout);
		printed = 1;
	}

	/* Do SOA RNAME. */
	n = dn_expand(base, end, rdata, domain, sizeof domain);
	CHK(n, "dn_expand");
	rdata += n;
	if (ISSET(soa_rname, fields)) {
		if (printed)
			fputc(' ', stdout);
		fputs(domain, stdout);
		printed = 1;
	}

	/* Check length all at once. */
	if (end - rdata < 5 * NS_INT32SZ) {
		fprintf(stderr, "soa rdlength too short\n");
		return (1);
	}

	/* Do SOA SERIAL. */
	l = ns_get32(rdata); rdata += NS_INT32SZ;
	if (ISSET(soa_serial, fields)) {
		fprintf(stdout, "%s%lu", printed ? " " : "", l);
		printed = 1;
	}

	/* Do SOA REFRESH. */
	l = ns_get32(rdata); rdata += NS_INT32SZ;
	if (ISSET(soa_refresh, fields)) {
		fprintf(stdout, "%s%lu", printed ? " " : "", l);
		printed = 1;
	}

	/* Do SOA RETRY. */
	l = ns_get32(rdata); rdata += NS_INT32SZ;
	if (ISSET(soa_retry, fields)) {
		fprintf(stdout, "%s%lu", printed ? " " : "", l);
		printed = 1;
	}

	/* Do SOA EXPIRE. */
	l = ns_get32(rdata); rdata += NS_INT32SZ;
	if (ISSET(soa_expire, fields)) {
		fprintf(stdout, "%s%lu", printed ? " " : "", l);
		printed = 1;
	}

	/* Do SOA MINIMUM. */
	l = ns_get32(rdata); rdata += NS_INT32SZ;
	if (ISSET(soa_minimum, fields)) {
		fprintf(stdout, "%s%lu", printed ? " " : "", l);
		printed = 1;
	}

	return (0);
}

/* Private - NS. */

static int
ns_get(ns_msg *msg, ns_rr *rr, int fields) {
	const u_char *base = ns_msg_base(*msg);
	const u_char *end = ns_msg_end(*msg);
	const u_char *rdata = ns_rr_rdata(*rr);
	char domain[NS_MAXDNAME];
	int n;

	/* Do NS NSDNAME. */
	n = dn_expand(base, end, rdata, domain, sizeof domain);
	CHK(n, "dn_expand");
	rdata += n;
	if (ISSET(ns_nsdname, fields))
		fputs(domain, stdout);

	return (0);
}

/* Private - A. */

static int
a_get(ns_msg *msg, ns_rr *rr, int fields) {
	const u_char *end = ns_msg_end(*msg);
	const u_char *rdata = ns_rr_rdata(*rr);
	struct in_addr addr;

	/* Do A ADDRESS. */
	if (end - rdata < NS_INT32SZ) {
		fprintf(stderr, "a rdlength too short\n");
		return (1);
	}
	addr.s_addr = htonl(ns_get32(rdata)); rdata += NS_INT32SZ;
	if (ISSET(a_address, fields))
		printf("%s", inet_ntoa(addr));

	return (0);
}

/* Private - AAAA. */

static int
aaaa_get(ns_msg *msg, ns_rr *rr, int fields) {
	const u_char *end = ns_msg_end(*msg);
	const u_char *rdata = ns_rr_rdata(*rr);
	const void *addr;

	/* Do AAAA ADDRESS. */
	if (end - rdata < 16) {
		fprintf(stderr, "a rdlength too short\n");
		return (1);
	}
	addr = rdata;  rdata += 16;
	if (ISSET(a_address, fields)) {
		char t[sizeof "ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255"];
		(void) inet_ntop(AF_INET6, addr, t, sizeof t);
		printf("%s", t);
	}

	return (0);
}

/* Private - MX. */

static int
mx_get(ns_msg *msg, ns_rr *rr, int fields) {
	const u_char *base = ns_msg_base(*msg);
	const u_char *end = ns_msg_end(*msg);
	const u_char *rdata = ns_rr_rdata(*rr);
	char domain[NS_MAXDNAME];
	int n, printed = 0;
	u_long l;

	/* Do MX PREFERENCE. */
	if (end - rdata < NS_INT16SZ) {
		fprintf(stderr, "mx rdlength too short\n");
		return (1);
	}
	l = ns_get16(rdata); rdata += NS_INT16SZ;
	if (ISSET(mx_preference, fields))
		printf("%s%lu", printed++ ? " " : "", l);

	/* Do MX EXCHANGE. */
	n = dn_expand(base, end, rdata, domain, sizeof domain);
	CHK(n, "dn_expand");
	rdata += n;
	if (ISSET(mx_exchange, fields))
		printf("%s%s", printed++ ? " " : "", domain);

	return (0);
}
